from django.db import models
import numpy as np
from ..annotation import ImageClassificationAnnotation

def find_segments_in_prediction_array(prediction_array: np.array, min_frame_len: int):
    """
    Expects a prediction array of shape (num_frames) and a minimum frame length.
    Returns a list of tuples (start_frame_number, end_frame_number) that represent the segments.
    """
    # Add False to the beginning and end to detect changes at the array boundaries
    padded_prediction = np.pad(prediction_array, (1, 1), 'constant', constant_values=False)
    
    # Find the start points and end points of the segments
    diffs = np.diff(padded_prediction.astype(int))
    segment_starts = np.where(diffs == 1)[0]
    segment_ends = np.where(diffs == -1)[0]
    
    # Filter segments based on min_frame_len
    segments = [(start, end) for start, end in zip(segment_starts, segment_ends) if end - start >= min_frame_len]
    
    return segments

class AbstractLabelVideoSegment(models.Model):
    video = None # Placeholder for the video field, to be defined in derived classes
    prediction_meta = None # Placeholder for the prediction_meta field, to be defined in derived classes
    start_frame_number = models.IntegerField()
    end_frame_number = models.IntegerField()
    source = models.ForeignKey("InformationSource", on_delete=models.CASCADE)
    label = models.ForeignKey("Label", on_delete=models.CASCADE)

    class Meta:
        abstract = True

    def __str__(self):
        return self.video.file.path + " Label - " + self.label.name + " - " + str(self.start_frame_number) + " - " + str(self.end_frame_number)
    
    def get_frames(self):
        return self.video.get_frame_range(self.start_frame_number, self.end_frame_number)
    
    def get_annotations(self) -> ImageClassificationAnnotation.objects:
        frames = self.get_frames()
        annotations = ImageClassificationAnnotation.objects.filter(frame__in=frames, label=self.label)

        return annotations
    
    def get_frame_without_annotation(self):
        """
        Get a frame without an annotation.
        """
        assert 1 == 2, "This method should be overridden in derived classes"

    def get_segment_len_in_s(self):
        return (self.end_frame_number - self.start_frame_number) / self.video.fps

class LegacyLabelVideoSegment(AbstractLabelVideoSegment):
    video = models.ForeignKey("LegacyVideo", on_delete=models.CASCADE)
    prediction_meta = models.ForeignKey("LegacyVideoPredictionMeta", on_delete=models.CASCADE, related_name="video_segments")

    def get_annotations(self) -> ImageClassificationAnnotation.objects:
        frames = self.get_frames()
        annotations = ImageClassificationAnnotation.objects.filter(legacy_frame__in=frames, label=self.label)

        return annotations

    def get_frames_without_annotation(self, n_frames):
        """
        Get a frame without an annotation.
        """
        frames = self.get_frames()
        annotations = ImageClassificationAnnotation.objects.filter(legacy_frame__in=frames, label=self.label)

        annotated_frames = [annotation.legacy_frame for annotation in annotations]
        frames_without_annotation = [frame for frame in frames if frame not in annotated_frames]

        # draw n random frames
        if len(frames_without_annotation) > n_frames:
            frames_without_annotation = np.random.choice(frames_without_annotation, n_frames, replace=False)

        return frames_without_annotation


from .unit import Unit
from .information_source import InformationSource
from .center import Center
from .persons import (
    Person,
    Patient, PatientForm, PatientSerializer,
    Examiner, ExaminerSerializer,
)

from .examination import (
    Examination,
    ExaminationType,
    ExaminationTime,
    ExaminationTimeType,
)

from .data_file import *

from .patient_examination import PatientExamination

from .label import (
    Label,
    LabelType,
    LabelSet
)

from .annotation import (
    ImageClassificationAnnotation,
    LegacyBinaryClassificationAnnotationTask,
)

from .legacy_data import (
    LegacyImage,
)

from .ai_model import (
    ModelMeta,
    ModelType,
)

from .prediction import (
    ImageClassificationPrediction,
    LegacyVideoPredictionMeta
)

from .hardware import (
    EndoscopyProcessor,
    Endoscope
)


# command to fetch the legacy dataset and save it to a json file.
from django.core.management.base import BaseCommand, CommandError
from agl_base_db.queries import generate_legacy_dataset_output 
import json

LABELSET_NAME = "multilabel_classification"
LABELSET_VERSION = 0

class Command(BaseCommand):
    help = 'Fetches the legacy dataset and saves it to a jsonl file'

    def handle(self, *args, **options):
        output, labelset = generate_legacy_dataset_output(LABELSET_NAME, LABELSET_VERSION)
        # output is a list of dicts, each dict is an image
        labels_in_order = labelset.get_labels_in_order()

        # write to file
        with open('legacy_dataset.jsonl', 'w') as f:
            for img_dict in output:
                f.write(json.dumps(img_dict) + '\n')


        _dict = {
            "labels": [label.name for label in labels_in_order],
            "name": LABELSET_NAME,
            "version": LABELSET_VERSION
        }

        with open('legacy_dataset_info.json', 'w') as f:
            f.write(json.dumps(_dict))

        self.stdout.write(self.style.SUCCESS('Successfully fetched and saved the legacy dataset'))



from django.db import models

def get_prediction_information_source():
    _source = InformationSource.objects.get(name="prediction")

    # make sure to return only one object
    assert _source, "No prediction information source found"
    return _source

class InformationSourceManager(models.Manager):
    def get_by_natural_key(self, name):
        return self.get(name=name)

class InformationSource(models.Model):
    objects = InformationSourceManager()

    name = models.CharField(max_length=100)
    name_de = models.CharField(max_length=100, blank=True, null=True)
    name_en = models.CharField(max_length=100, blank=True, null=True)

    def __str__(self):
        return self.name



from django.core.management import BaseCommand, call_command

class Command(BaseCommand):
    help = 'Run all data loading commands in the correct order'

    def add_arguments(self, parser):
        parser.add_argument(
            '--verbose',
            action='store_true',
            help='Display verbose output for all commands',
        )

    def handle(self, *args, **options):
        verbose = options['verbose']

        self.stdout.write(self.style.SUCCESS("Populating base db models with data..."))

        # Run the load_endoscopy_processor_data command
        self.stdout.write(self.style.SUCCESS("Running load_endoscopy_processor_data..."))
        call_command('load_endoscopy_processor_data', verbose=verbose)

        # Run the load_unit_data command
        self.stdout.write(self.style.SUCCESS("Running load_unit_data..."))
        call_command('load_unit_data', verbose=verbose)

        # Run the load_information_source command
        self.stdout.write(self.style.SUCCESS("Running load_information_source..."))
        call_command('load_information_source', verbose=verbose)

        # Run the load_center_data command
        self.stdout.write(self.style.SUCCESS("Running load_center_data..."))
        call_command('load_center_data', verbose=verbose)

        # Run the load_examination_data command
        self.stdout.write(self.style.SUCCESS("Running load_examination_data..."))
        call_command('load_examination_data', verbose=verbose)

        # Run the load_label_data command
        self.stdout.write(self.style.SUCCESS("Running load_label_data..."))
        call_command('load_label_data', verbose=verbose)

        # Run the load_ai_model_data command
        self.stdout.write(self.style.SUCCESS("Running load_ai_model_data..."))
        call_command('load_ai_model_data', verbose=verbose)
        
        self.stdout.write(self.style.SUCCESS("All data loading commands executed successfully."))



# Django command to fetch all LegacyVideo objects, delete the associated file and delete the object from the database

from agl_base_db.models.data_file.video import LegacyVideo
from tqdm import tqdm
from django.core.management.base import BaseCommand

class Command(BaseCommand):
    """
    Deletes all LegacyVideos in the database.
    """
    help = 'Deletes all LegacyVideos in the database.'

    def handle(self, *args, **options):
        videos = LegacyVideo.objects.all()
        for video in tqdm(videos):
            video.file.delete()
            video.delete()


from pathlib import Path
from tkinter import E, END

DATA_DIR = Path(__file__).parent

INFORMATION_DATA_DIR = DATA_DIR / "information_source"
UNIT_DATA_DIR = DATA_DIR / "unit"
CENTER_DATA_DIR = DATA_DIR / "center"
EXAMINATION_DATA_DIR = DATA_DIR / "examination"
LABEL_DATA_DIR = DATA_DIR / "label"
MODEL_TYPE_DATA_DIR = DATA_DIR / "model_type"
ENDOSCOPY_PROCESSOR_DATA_DIR = DATA_DIR / "endoscopy_processor"


from agl_base_db.models.annotation.image_classification import ImageClassificationAnnotation
from agl_base_db.models.label.label import Label

from django.db import models


class AbstractFrame(models.Model):
    video = None # Placeholder for the video field, to be defined in derived classes
    frame_number = models.IntegerField()
    # Add any other fields you need to store frame-related information
    image = models.ImageField(upload_to="frames")  # Or some other field type, depending on how you're storing the frame
    suffix = models.CharField(max_length=255)
    # ImageClassificationAnnotation has a foreign key to this model (related name: image_classification_annotations)

    class Meta:
        # Ensure that for each video, the frame_number is unique
        abstract = True
        unique_together = ('video', 'frame_number')
        # Optimize for retrieval in frame_number order
        indexes = [models.Index(fields=['video', 'frame_number'])]


    def __str__(self):
        return self.video.file.path + " - " + str(self.frame_number)

    def get_frame_model(self):
        assert 1 == 2, "This method should be overridden in derived classes"

    def get_classification_annotations(self):
        """
        Get all image classification annotations for this frame.
        """
        return ImageClassificationAnnotation.objects.filter(frame=self)

    def get_classification_annotations_by_label(self, label:Label):
        """
        Get all image classification annotations for this frame with the given label.
        """
        return ImageClassificationAnnotation.objects.filter(frame=self, label=label)

    def get_classification_annotations_by_value(self, value:bool):
        """
        Get all image classification annotations for this frame with the given value.
        """
        return ImageClassificationAnnotation.objects.filter(frame=self, value=value)

    def get_classification_annotations_by_label_and_value(self, label:Label, value:bool):
        """
        Get all image classification annotations for this frame with the given label and value.
        """
        return ImageClassificationAnnotation.objects.filter(frame=self, label=label, value=value)


from django.db import models
from rest_framework import serializers

class ExaminationTimeTypeManager(models.Manager):
    def get_by_natural_key(self, name):
        return self.get(name=name)
    
class ExaminationTimeType(models.Model):
    objects = ExaminationTimeTypeManager()
    name = models.CharField(max_length=100, unique=True)
    name_de = models.CharField(max_length=100, blank=True, null=True)
    name_en = models.CharField(max_length=100, blank=True, null=True)
    examinations = models.ManyToManyField('Examination', blank=True)

    def __str__(self):
        return self.name
    
    def natural_key(self):
        return (self.name,)

    class Meta:
        verbose_name = 'Examination Time Type'
        verbose_name_plural = 'Examination Time Types'
        ordering = ['name']

class ExaminationTimeTypeSerializer(serializers.ModelSerializer):
    
    class Meta:
        model = ExaminationTimeType
        fields = '__all__'


from django.db import models

class ImageClassificationPrediction(models.Model):
    """
    A class representing an image classification prediction.

    Attributes:
        label (Label): The label of the prediction.
        frame (Frame): The frame of the prediction.
        confidence (float): The confidence of the prediction.
        date_created (datetime): The date the prediction was created.

    """
    label = models.ForeignKey("Label", on_delete=models.CASCADE, related_name="image_classification_predictions")
    frame = models.ForeignKey("Frame", on_delete=models.CASCADE, blank=True, null=True, related_name="image_classification_predictions")
    legacy_frame = models.ForeignKey("LegacyFrame", on_delete=models.CASCADE, blank=True, null=True, related_name="image_classification_predictions")
    legacy_image = models.ForeignKey("LegacyImage", on_delete=models.CASCADE, blank=True, null=True, related_name="image_classification_predictions")
    value = models.BooleanField()
    confidence = models.FloatField()
    model_meta = models.ForeignKey("ModelMeta", on_delete=models.CASCADE, related_name="image_classification_predictions")
    date_created = models.DateTimeField(auto_now_add=True)

    class Meta:
        unique_together = ('label', 'frame', 'model_meta')

    def get_image_object(self):
        """
        Get the image of the prediction.
        """
        if self.frame:
            return self.frame
        elif self.legacy_frame:
            return self.legacy_frame
        elif self.legacy_image:
            return self.legacy_image
        else:
            return None



from .person import Person
from django import forms
from django.forms import DateInput
from rest_framework import serializers
from ..patient_examination import PatientExamination
from ..data_file import ReportFile
from django.db import models

class Patient(Person):
    """
    A class representing a patient.

    Attributes inhereted from Person:
        first_name (str): The first name of the patient.
        last_name (str): The last name of the patient.
        dob (datetime.date): The date of birth of the patient.
        gender (str): The gender of the patient.
        email (str): The email address of the patient.
        phone (str): The phone number of the patient.

    """
    center = models.ForeignKey("Center", on_delete=models.CASCADE, blank=True, null=True)

    def __str__(self):
        return self.first_name + " " + self.last_name + " (" + str(self.dob) + ")"
    
    def get_unmatched_report_files(self): #field: self.report_files; filter: report_file.patient_examination = None
        '''Returns all report files for this patient that are not matched to a patient examination.'''

        return self.report_files.filter(patient_examination=None)

    def get_unmatched_video_files(self): #field: self.videos; filter: video.patient_examination = None
        '''Returns all video files for this patient that are not matched to a patient examination.'''
        return self.videos.filter(patient_examination=None)

    def get_patient_examinations(self): #field: self.patient_examinations
        '''Returns all patient examinations for this patient ordered by date (most recent is first).'''
        return self.patient_examinations.order_by('-date')
    
    def create_examination_by_report_file(self, report_file:ReportFile):
        '''Creates a patient examination for this patient based on the given report file.'''
        patient_examination = PatientExamination(patient=self, report_file=report_file)
        patient_examination.save()
        return patient_examination
        
    
class PatientSerializer(serializers.ModelSerializer):
    class Meta:
        model = Patient
        fields = '__all__'

class PatientForm(forms.ModelForm):
    class Meta:
        model = Patient
        fields = '__all__'
        widgets = {
            'dob': DateInput(attrs={'type': 'date'}),
        }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        for field in self.fields.values():
            field.widget.attrs['class'] = 'form-control'



import os
import yaml

def load_model_data_from_yaml(
        command,
        model_name,
        metadata,
        verbose,
    ):

    if verbose:
        command.stdout.write(f"Start Loading {model_name}")
    model = metadata["model"]
    dir = metadata["dir"]
    foreign_keys = metadata["foreign_keys"]
    foreign_key_models = metadata["foreign_key_models"]

    for file in [f for f in os.listdir(dir) if f.endswith('.yaml')]:
        with open(os.path.join(dir, file), 'r') as f:
            yaml_data = yaml.safe_load(f)
        
        load_data_with_foreign_keys(
            command,
            model,
            yaml_data,
            foreign_keys,
            foreign_key_models,
            verbose
        )

def load_data_with_foreign_keys(command, model, yaml_data, foreign_keys, foreign_key_models, verbose):
            # Since pathology types is a ManyToMany field, we need to hack arount
            for entry in yaml_data:
                fields = entry.get('fields', {})
                name = fields.pop('name', None)
                many_to_many_tuples = []
                foreign_key_tuples = zip(foreign_keys, foreign_key_models)
                for foreign_key, foreign_key_model in foreign_key_tuples:
                    target_natural_key = fields.pop(foreign_key, None)
   

                    if isinstance(target_natural_key, list):
                        # the field is a Many to X field.
                        fk_objects = [foreign_key_model.objects.get_by_natural_key(_) for _ in target_natural_key]
                        fk_tuple = (foreign_key, fk_objects)
                        many_to_many_tuples.append(fk_tuple)
                        continue
                    # Use the natural key to look up the related object
                    try:
                        obj = foreign_key_model.objects.get_by_natural_key(target_natural_key)
                    except model.DoesNotExist:
                        command.stderr.write(command.style.ERROR(f'{model.__name__} with natural key {target_natural_key} does not exist. Skipping {name}.'))
                        raise Exception(f'{model.__name__} with natural key {target_natural_key} does not exist. Skipping {name}.')

                    # Assign the related object to the field
                    fields[foreign_key] = obj

                if name:
                    obj, created = model.objects.get_or_create(name=name, defaults=fields)
                else:
                    obj, created = model.objects.get_or_create(**fields)
                if many_to_many_tuples:
                    for fk, fk_objects in many_to_many_tuples:
                        getattr(obj, fk).set(fk_objects)

                if created and verbose:
                    command.stdout.write(command.style.SUCCESS(f'Created {model.__name__} {name}'))
                elif verbose:
                    command.stdout.write(command.style.WARNING(f'Skipped {model.__name__} {name}, already exists'))



from django.db import models
from rest_framework import serializers

class ExaminationManager(models.Manager):
    def get_by_natural_key(self, name):
        return self.get(name=name)
    
class Examination(models.Model):
    name = models.CharField(max_length=100, unique=True)
    name_de = models.CharField(max_length=100, blank=True, null=True)
    name_en = models.CharField(max_length=100, blank=True, null=True)
    examination_types = models.ManyToManyField('ExaminationType', blank=True)
    objects = ExaminationManager()
    date = models.DateField(blank=True, null=True)
    time = models.TimeField(blank=True, null=True)

    def __str__(self):
        return self.name
    
    def natural_key(self):
        return (self.name,)

    class Meta:
        verbose_name = 'Examination'
        verbose_name_plural = 'Examinations'
        ordering = ['name']

class ExaminationSerializer(serializers.ModelSerializer): 
    class Meta:
        model = Examination
        fields = '__all__'



from django.db import models

class VideoImportMeta(models.Model):
    processor = models.ForeignKey('EndoscopyProcessor', on_delete=models.CASCADE)
    endoscope = models.ForeignKey('Endoscope', on_delete=models.CASCADE, blank=True, null=True)
    center = models.ForeignKey('Center', on_delete=models.CASCADE)
    video_anonymized = models.BooleanField(default=False)
    video_patient_data_detected = models.BooleanField(default=False)
    outside_detected = models.BooleanField(default=False)
    patient_data_removed = models.BooleanField(default=False)
    outside_removed = models.BooleanField(default=False)
    



from agl_base_db.models.annotation.image_classification import ImageClassificationAnnotation
from agl_base_db.models.label.label import Label
from .base_classes import AbstractFrame
from django.db import models

class Frame(AbstractFrame):
    video = models.ForeignKey("Video", on_delete=models.CASCADE, related_name="frames")


class LegacyFrame(AbstractFrame):
    video = models.ForeignKey("LegacyVideo", on_delete=models.CASCADE, related_name='frames')
    frame_number = models.IntegerField()
    image = models.ImageField(upload_to="legacy_frames", blank=True, null=True)
    suffix = models.CharField(max_length=255)
    # ImageClassificationAnnotation has a foreign key to this model (related name: image_classification_annotations)

    class Meta:
        unique_together = ('video', 'frame_number')
        indexes = [
            models.Index(fields=['video', 'frame_number']),
        ]

    def get_classification_annotations(self):
        """
        Get all image classification annotations for this frame.
        """
        return ImageClassificationAnnotation.objects.filter(legacy_frame=self)

    def get_classification_annotations_by_label(self, label:Label):
        """
        Get all image classification annotations for this frame with the given label.
        """
        return ImageClassificationAnnotation.objects.filter(legacy_frame=self, label=label)

    def get_classification_annotations_by_value(self, value:bool):
        """
        Get all image classification annotations for this frame with the given value.
        """
        return ImageClassificationAnnotation.objects.filter(legacy_frame=self, value=value)

    def get_classification_annotations_by_label_and_value(self, label:Label, value:bool):
        """
        Get all image classification annotations for this frame with the given label and value.
        """
        return ImageClassificationAnnotation.objects.filter(legacy_frame=self, label=label, value=value)

    



# django command to delete all LegacyImage objects

from django.core.management.base import BaseCommand, CommandError
from agl_base_db.models import LegacyImage
from tqdm import tqdm

class Command(BaseCommand):
    help = 'Deletes all LegacyImage objects including the linked files (legacy_image.image)'

    def handle(self, *args, **options):
        legacy_images = LegacyImage.objects.all()

        for legacy_image in tqdm(legacy_images):
            legacy_image.image.delete()
            legacy_image.delete()

        self.stdout.write(self.style.SUCCESS('Successfully deleted all LegacyImage objects and linked image files'))





# Django Command to:
# fetch all LegacyVideos and extract frames from them 
# track progress for all videos and for each video with tqdm

from agl_base_db.models.data_file.video import LegacyVideo
from tqdm import tqdm
from django.core.management.base import BaseCommand

class Command(BaseCommand):
    """
    Extracts frames from all LegacyVideos in the database.
    """
    help = 'Extracts frames from all LegacyVideos in the database.'

    def handle(self, *args, **options):
        videos = LegacyVideo.objects.all()
        for video in tqdm(videos):
            video.extract_all_frames()
